from collections import defaultdict

import numpy as np

from rl_agents.agents.functions.base import BaseQFunction


class QMatrixFunction(BaseQFunction):
    """A simple Q-table using numpy array.

    Parameters
    ----------
    n_states : int
        Number of states in the state space.
    n_actions : int
        Number of actions in the action space.
    method : str
        Initialization method. Possible values: 'zeros', 'random' or 'ones'.

    Attributes
    ----------
    q_table : numpy.ndarray(float, ndims=2)
        Q-Table matrix, rows are the states and columns the actions.

    """

    def __init__(self, n_states, n_actions, method="zeros"):
        # if method not in ["zeros", "ones", "random"]:

        if method == "zeros":
            self.q_table = np.zeros((n_states, n_actions))
        elif method == "random":
            self.q_table = np.random.random((n_states, n_actions))
        elif method == "ones":
            self.q_table = np.ones((n_states, n_actions))
        else:
            raise ValueError(
                "Invalid Method, options: 'zeros', 'random' or 'ones'. "
            )

    def __call__(self, state, action):
        return self.q_table[state, action]

    def update(self, state, action, target):
        self.q_table[state, action] = target

    def get_values(self, state):
        return self.q_table[state, :]


class QTableFunction(BaseQFunction):
    """A simple Q-table using `defaultdict`.

    Parameters
    ----------
    n_states : int
        Number of states in the state space.
    n_actions : int
        Number of actions in the action space.
    method : str
        Initialization method. Possible values: 'zeros', 'random' or 'ones'.

    Attributes
    ----------
    q_table : defaultdict
        Each key is a vector representing the Q-values
        of each action for that state.

    """

    def __init__(self, n_states, n_actions, method="zeros"):
        if method == "zeros":
            self.q_table = defaultdict(lambda: np.zeros(n_actions))
        elif method == "random":
            self.q_table = defaultdict(lambda: np.random.random(n_actions))
        elif method == "ones":
            self.q_table = defaultdict(lambda: np.ones(n_actions))
        else:
            raise ValueError(
                "Invalid Method, options: 'zeros', 'random' or 'ones'. "
            )

    def __call__(self, state, action):
        return self.q_table[state][action]

    def update(self, state, action, target):
        self.q_table[state][action] = target

    def get_values(self, state):
        return self.q_table[state]



import numpy as np


def simple_mab_runner(env, agent, n_trials):
    """Run a simple MAB experiment.

    Parameters
    ----------
    env : gym.Env
        Environment.
    agent : rl_agents.agents.mab.BaseMAB
        MAB agent.
    n_trials : int
        Number of trials until the experiment ends.

    Returns
    -------
    rewards : numpy.ndarray(float, ndims=1)
        Vector with the observed rewards.
    regrets : numpy.ndarray(float, ndims=1)
        Vector with the regret of each trial.
    optimal : numpy.ndarray(float, ndims=1)
        Vector containing the percentage of optimal actions taken.

    """
    _ = env.reset()
    regrets = np.zeros(n_trials)
    rewards = np.zeros(n_trials)
    optimals = np.zeros(n_trials)
    for ii in range(n_trials):
        arm_idx = agent.predict()
        _, reward, _, info = env.step(arm_idx)
        agent.learn(arm_idx, reward)
        rewards[ii] = (rewards.sum() + reward) / (ii + 1)
        regrets[ii] = (regrets.sum() + info["Regret"]) / (ii + 1)
        optimals[ii] = (optimals.sum() + info["Optimal"]) / (ii + 1)
    return rewards, regrets, optimals



import numpy as np

from rl_agents.agents.mab.base import BaseMAB


class Pursuit(BaseMAB):
    """Short summary.

    Parameters
    ----------
    n_arms : type
        Description of parameter `n_arms`.
    beta : type
        Description of parameter `beta`.

    Attributes
    ----------
    means : type
        Description of attribute `means`.
    p_arms : type
        Description of attribute `p_arms`.
    trials : type
        Description of attribute `trials`.
    n_arms
    beta

    """

    def __init__(self, n_arms, beta):
        self.n_arms = n_arms
        self.means = np.zeros(self.n_arms)
        self.beta = beta
        self.p_arms = np.ones(self.n_arms) / self.n_arms
        self.trials = np.zeros(self.n_arms)

    def learn(self, a_idx, reward):
        """Short summary.

        Parameters
        ----------
        reward : type
            Description of parameter `reward`.
        a_idx : type
            Description of parameter `a_idx`.

        Returns
        -------
        type
            Description of returned object.

        """
        self.means[a_idx] = (
            (self.means[a_idx] * self.trials[a_idx]) + reward
        ) / (self.trials[a_idx] + 1)
        self.trials[a_idx] += 1  # add trial
        ii = np.argmax(self.means)
        p_ii = self.p_arms[ii]
        self.p_arms = (1 - self.beta) * self.p_arms
        self.p_arms[ii] = p_ii + self.beta * (1 - p_ii)

    def predict(self):
        """Short summary.

        Returns
        -------
        type
            Description of returned object.

        """
        return np.random.choice(range(self.n_arms), p=self.p_arms)



import numpy as np
from tqdm import tqdm


def simple_tab_runner(env, agent, n_episodes):
    """Short summary.

    Parameters
    ----------
    env : gym.Env
        Environment.
    agent : rl_agents.agents.BaseAgent
        MAB agent.
    n_episodes : int
        Number of episodes to run.

    Returns
    -------
    rewards : numpy.ndarray(float, ndims=1)
        Vector with the total episode reward.

    """
    rewards = np.zeros(n_episodes)
    for ii in tqdm(range(n_episodes)):
        # Run episode:
        done = False
        episode_reward = 0
        obs = env.reset()
        while not done:
            action = agent.predict(obs)
            next_obs, reward, done, info = env.step(action)
            agent.learn(obs, action, reward, next_obs)
            agent.policy.update()
            obs = next_obs
            episode_reward += reward
        rewards[ii] = episode_reward
    return rewards



"""
The :mod:`rl_agents.agents.tabular` submodule includes:

* Q-Learning
* SARSA
* Expected SARSA
"""
from rl_agents.agents.tabular.td_learning import (  # isort:skip
    QLearningAgent,
    SarsaAgent,
    ExpectedSarsaAgent,
)

__all__: ["QLearningAgent", "SarsaAgent", "ExpectedSarsaAgent"]



import numpy as np

from rl_agents.agents.mab.base import BaseMAB


class EpsilonGreedy(BaseMAB):
    r"""Epsilon-Greedy agent.

    The agent uses the epsilon-greedy approach to solve the Multi-Armed
    bandit problem.

    The parameter :math:`\epsilon` is used for the
    exploration-exploitation trade-off. With probability :math:`\epsilon`
    the agent selects a random action, otherwise it selects the action
    that has the best average reward.

    Parameters
    ----------
    n_arms : int
        Number of actions (arms) of the MAB.
    epsilon : float
        Probability of selecting a random action.

    Attributes
    ----------
    means : numpy.array(float, ndim=1)
        Vector containing the average reward of each arm.
    trials : numpy.array(float, ndim=1)
        Vector containing the number of trials made to each arm.

    """

    def __init__(self, n_arms, epsilon):
        self.epsilon = epsilon
        self.n_arms = n_arms
        self.means = np.zeros(self.n_arms)
        self.trials = np.zeros(self.n_arms)

    def learn(self, a_idx, reward):
        """Make `EpsilonGreedy` agent learn from the interaction.

        The `EpsilonGreedy` agent learns from its previous choice
        and the reward received from this action.
        Updates the means and the trials.

        Parameters
        ----------
        reward : float
            Reward received from the system after taking action a_idx.
        a_idx : int
            Index of the arm pulled (action taken).

        """
        self.means[a_idx] = (
            self.means[a_idx] * self.trials[a_idx] + reward
        ) / (self.trials[a_idx] + 1)
        self.trials[a_idx] += 1  # add trial

    def predict(self):
        r"""Predict next action.

        With probability :math:`\epsilon` the agent selects a random arm.
        With probability :math:`1 - \epsilon` the agent selects the arm that
        has the best average reward.

        Returns
        -------
        int
            Index of chosen action.

        """
        if np.random.rand() < self.epsilon:
            a = np.random.randint(low=0, high=self.n_arms)
        else:
            a = self.means.argmax()
        return a


class DecayEpsilon(BaseMAB):
    r"""Agent that follows an epsilon-decreasing policy.

    The agent uses the epsilon-greedy approach to solve the Multi-Armed
    bandit problem, but with a decay in the epsilon.

    The parameter :math:`\epsilon` is used for the exploration-exploitation
    trade-off. With probability :math:`\epsilon` the agent selects a
    random action, otherwise it selects the action that has the
    best average reward. After each interaction the epsilon is updated as
    epsilon = epsilon * decay.

    Parameters
    ----------
    n_arms : int
        Number of actions (arms) of the MAB.
    max_epsilon : float
        Initial epsilon.
    decay : float
        Decay of the epsilon.

    Attributes
    ----------
    epsilon : float
        Epsilon of the agent. Constantly updated as epsilon = epsilon*decay
    means : numpy.array(float, ndim=1)
        Vector containing the average reward of each arm.
    trials : numpy.array(float, ndim=1)
        Vector containing the number of trials made to each arm.

    """

    def __init__(self, n_arms, max_epsilon, decay):
        self.epsilon = max_epsilon
        self.n_arms = n_arms
        self.means = np.zeros(self.n_arms)
        self.trials = np.zeros(self.n_arms)
        self.decay = decay

    def learn(self, a_idx, reward):
        """Make the `DecayEpsilon` agent learn from the interaction.

        The MAB agent learns from its previous choice and the reward received
        from this action. Updates the means and the trials.

        Parameters
        ----------
        reward : float
            Reward received from the system after taking action a_idx.
        a_idx : int
            Index of the arm pulled (action taken).

        """
        self.means[a_idx] = (
            self.means[a_idx] * self.trials[a_idx] + reward
        ) / (self.trials[a_idx] + 1)
        self.trials[a_idx] += 1  # add trial

    def predict(self):
        r"""Predict next action and update epsilon.

        With probability :math:`\epsilon` the agent selects a random arm.
        With probability :math:`1 - \epsilon` the agent selects the arm that
        has the best average reward.

        Returns
        -------
        int
            Index of chosen action.

        """
        if np.random.rand() < self.epsilon:
            a_idx = np.random.randint(low=0, high=self.n_arms)
        else:
            a_idx = self.means.argmax()
        self.epsilon = self.epsilon * self.decay
        return a_idx



from abc import ABC, abstractmethod

import numpy as np


class BasePolicy(ABC):
    """
    A basic policy for tabular agents.

    The policy is a function that maps a state to an action.

    """

    @abstractmethod
    def __call__(self, q_values):
        """Select an action for the current timestep.

        This method returns the action selected by the current policy

        Parameters
        ----------
        q_values : numpy.ndarray(float, ndim=1)
            Q-value of each action.

        Returns
        -------
        int
            Index of chosen action.

        """

    @abstractmethod
    def update(self):
        """Update the policy.

        """

    @abstractmethod
    def get_values(self, q_values):
        """Return the probabilities associated with each action.

        Parameters
        ----------
        q_values : numpy.ndarray(float, ndim=1)
            Q-value of each action.

        Returns
        -------
        numpy.ndarray(float, ndim=1)
            Probabilities associated with each action.

        """


class EGreedyPolicy(BasePolicy):
    """Epsilon-greedy policy for tabular agents.

    Parameters
    ----------
    epsilon : float
        Epsilon parameter to control the exploration-exploitation trade-off.

    Attributes
    ----------
    epsilon : float
        Exploration-exploitation parameter.

    """

    def __init__(self, epsilon):
        """Instantiate a `EGreedyPolicy` object.

        Parameters
        ----------
        epsilon : float
            Epsilon parameter to control the
            exploration-exploitation trade-off.


        """
        # Check for valid values:
        if (epsilon >= 1) or (epsilon <= 0):
            raise ValueError("Invalid value for epsilon, 0 <= epsilon <= 1")
        self.epsilon = epsilon

    def __call__(self, q_values):
        r"""Select an action based on the :math:`\epsilon`-greedy policy.


        Parameters
        ----------
        q_values : numpy.ndarray(float, ndim=1)
            Line from the Q-Table, corresponding to Q-values
            for the chosen state.

        Returns
        -------
        int
            Chosen action index

        """
        # Explorarion case:
        if np.random.rand() < self.epsilon:
            a_idx = np.random.randint(low=0, high=q_values.size)
        # Exploitation case:
        else:
            a_idx = np.argmax(q_values)
        return a_idx

    def update(self):
        pass

    def get_values(self, q_values):
        # Probababilites of exploration:
        output = np.ones(q_values.size) * self.epsilon / q_values.size
        # Add the exploitation probability:
        output[q_values.argmax()] += 1 - self.epsilon
        return output


class EDecreasePolicy(EGreedyPolicy):
    """Decreasing epsilon-greedy policy for tabular agents.

    Parameters
    ----------
    epsilon : float
        Epsilon parameter to control the exploration-exploitation trade-off.
    epsilon_min : float
        Minimum epsilon acceptable
    decay : float
        Decay for epsilon: epsilon <- epsilon * decay

    Attributes
    ----------
    epsilon_min : float
    decay : float

    """

    def __init__(self, epsilon, epsilon_min, decay):
        super().__init__(epsilon)
        self.epsilon_min = epsilon_min
        self.decay = decay

    # The call method is the same as the epsilon-greedy

    def update(self):
        if self.epsilon > self.epsilon_min:
            if self.epsilon * self.decay > self.epsilon_min:
                self.epsilon = self.epsilon * self.decay
            else:
                self.epsilon = self.epsilon_min

    # The get_values method is the same as the epsilon-greedy


class BoltzmanPolicy(BasePolicy):
    r"""Boltzman(Softmax) policy for tabular agents.

    .. math::
        \begin{equation}
        \pi\left(s, a\right) = \frac{e^{Q\left(s, a\right) / T}}
                               {\sum_{i=1}^{m} e^{Q\left(s, a_i\right)/ T}}
        \end{equation}

    Attributes
    ----------
    temperature : float
        Temperature parameter to control the
        exploration-exploitation trade-off.

    """

    def __init__(self, temperature):
        """Instantiate a `BoltzmanPolicy` object.

        Parameters
        ----------
        temperature : float
            Temperature parameter to control the
            exploration-exploitation trade-off.


        """
        # Check for valid values:
        if temperature < 0:
            raise ValueError("Invalid temperature value, T >= 0")
        self.temperature = temperature

    def __call__(self, q_values):
        r"""Select an action based on the Boltzman policy.

        .. math::
            \pi\left(s, a\right) = \frac{e^{Q\left(s, a\right) / T}}
                                   {\sum_{i=1}^{m} e^{Q\left(s, a_i\right)/ T}}

        Parameters
        ----------
        q_values : numpy.ndarray(float, ndim=1)
            Line from the Q-Table, corresponding to Q-values
            for the chosen state.

        Returns
        -------
        int
            Chosen action index

        """
        # Exponential:
        e_x = np.exp(q_values / self.temperature)
        # Probabilities
        p_arms = e_x / e_x.sum()
        return np.random.choice(range(q_values.size), p=p_arms)

    def update(self):
        pass

    def get_values(self, q_values):
        e_x = np.exp(q_values / self.temperature)
        p_arms = e_x / e_x.sum()
        return p_arms



import numpy as np

from rl_agents.agents.core import BaseAgent
from rl_agents.agents.functions import QMatrixFunction
from rl_agents.agents.policies import EGreedyPolicy


class TDAgent(BaseAgent):
    """A base Temporal-Difference Agent.

    This agent is used to build the basic TD algorithms:
        * Q-Learning
        * SARSA
        * Expected-SARSA

    Parameters
    ----------
    n_states : int
        Number of states in the state space.
    n_actions : int
        Number of actions in the action space.
    alpha : float
        Learning rate.
    gamma : float
        Discount factor.
    policy : rl_agents.agents.policies.base.BasePolicy
        A policy object.
    q_function : rl_agents.agents.functions.base.BaseQFunction
        A Q-Function class.

    Attributes
    ----------
    n_states : int
        Number of states.
    n_actions : int
        Number of actions.
    alpha : float
        Learning rate.
    gamma : float
        Discount factor.
    policy : BasePolicy
        Policy instance
    q_function : BaseQFunction
        Qfunction instance

    """

    def __init__(
        self,
        n_states,
        n_actions,
        alpha,
        gamma,
        policy=EGreedyPolicy(0.1),
        q_function=QMatrixFunction,
        q_func_kwargs=None,
    ):
        self.n_states = n_states
        self.n_actions = n_actions
        self.policy = policy
        q_func_kwargs = {} if q_func_kwargs is None else q_func_kwargs.copy()
        self.q_function = q_function(self.n_states, self.n_actions, **q_func_kwargs)
        self.alpha = alpha
        self.gamma = gamma

    def predict(self, state, eval=False):
        """Predict the next action the agent should take.

        Parameters
        ----------
        state : int
            Index of the state.
        eval : bool
            Flag to indicate if the agent is in a test setting (evaluation)

        Returns
        -------
        int
            Action index to be taken.

        """
        # Get the Q-Values associated with that state:
        q_values = self.q_function.get_values(state)
        # Utilize the policy:
        if eval:
            action = np.argmax(q_values)
        else:
            action = self.policy(q_values)
        return action

    def learn(self, state, action, reward, next_state):
        r"""Learn from the interaction.

        TD update:

        .. math::
            Q(s,a) \leftarrow (1-\alpha) Q(s,a) + \alpha(r + \gamma F(s))

        The difference between the TD algorithms
        (Q-Learning, SARSA, Expected-SARSA) comes from the difference
        in the :math:`F(s)` function.

        Parameters
        ----------
        state : type
            State in which the action was taken.
        action : type
            Action taken
        reward : float
            Reward received by the transition.
        next_state : type
            Next state the environment transitions.


        """
        # TD update Q(s,a) <- (1-alpha)Q(s,a) + alpha(r + gamma F(s))
        update = self.alpha * (reward + self.gamma * self._next_value(next_state))
        target_q = (1 - self.alpha) * self.q_function(state, action) + update
        # Update the Q-Function with the target:
        self.q_function.update(state, action, target_q)

    def _next_value(self, next_state):
        raise NotImplementedError


class QLearningAgent(TDAgent):
    r"""A Simple Q-Learning Agent

    Refer to `TDAgent` for reference on the parameters and methods.

    The Q-Learning update is given as:

    .. math::
        Q(s, a) \leftarrow  (1-\alpha) Q(s, a) +
                            \alpha \left[ r +
                            \gamma \max_{a' \in A} Q(s', a') \right]

    """

    def __init__(
        self,
        n_states,
        n_actions,
        alpha,
        gamma,
        policy=EGreedyPolicy(0.1),
        q_function=QMatrixFunction,
        q_func_kwargs=None,
    ):

        super().__init__(
            n_states, n_actions, alpha, gamma, policy, q_function, q_func_kwargs,
        )

    def _next_value(self, next_state):
        q_values = self.q_function.get_values(next_state)
        return q_values.max()


class SarsaAgent(TDAgent):
    r"""A Simple SARSA Agent

    Refer to `TDAgent` for reference on the parameters and methods.

    The SARSA update is given as:

    .. math::
        Q(s_t, a_t) \leftarrow  (1-\alpha) Q(s, a) +
        \alpha \left[ r +
        \gamma Q(s',\pi(s') ) \right]

    Where :math:`\pi(s)` returns the action chosen by the policy.

    """

    def __init__(
        self,
        n_states,
        n_actions,
        alpha,
        gamma,
        policy=EGreedyPolicy(0.1),
        q_function=QMatrixFunction,
        q_func_kwargs=None,
    ):

        super().__init__(
            n_states, n_actions, alpha, gamma, policy, q_function, q_func_kwargs,
        )
        self.next_action = None

    def predict(self, state, eval=False):
        """Predict the next action the SARSA agent should take.

        Parameters
        ----------
        state : int
            Index of the state.
        eval : bool
            Flag to indicate if the agent is in a test setting (evaluation)

        Returns
        -------
        int
            Action index to be taken.

        """
        # Get the Q-Values associated with that state:
        q_values = self.q_function.get_values(state)
        # Utilize the policy:
        if eval:
            action = np.argmax(q_values)
        else:
            if self.next_action == None:
                action = self.policy(q_values)
            else:
                action = self.next_action
        return action

    def _next_value(self, next_state):
        q_values = self.q_function.get_values(next_state)
        action = self.policy(q_values)
        self.next_action = action
        return self.q_function(next_state, action)


class ExpectedSarsaAgent(TDAgent):
    r"""A Simple expeted-SARSA Agent

    The Expeted-SARSA update is given as:

    .. math::
        Q(s_t, a_t) \leftarrow  (1-\alpha) Q(s, a) +
        \alpha \left[ r +
        \gamma \sum_{a' \in A} \pi(s',a') Q(s',a') \right]

    Where :math:`\pi(s,a)` returns the probability of taking
    action :math:`a` in state :math:`s`.
    """

    def __init__(
        self,
        n_states,
        n_actions,
        alpha,
        gamma,
        policy=EGreedyPolicy(0.1),
        q_function=QMatrixFunction,
        q_func_kwargs=None,
    ):

        super().__init__(
            n_states, n_actions, alpha, gamma, policy, q_function, q_func_kwargs,
        )

    def _next_value(self, next_state):
        q_values = self.q_function.get_values(next_state)
        pi_values = self.policy.get_values(q_values)
        return np.sum(q_values * pi_values)



"""
The :mod:`rl_agents.agents.mab` submodule includes:

* Epsilon Greedy
* Decreasing Epsilon Greedy
* UCBs: UCB, UCB1, UCB2
* Softmax
* Pursuit
"""


from rl_agents.agents.mab.egreedy import DecayEpsilon  # noqa: F401
from rl_agents.agents.mab.egreedy import EpsilonGreedy  # noqa: F401
from rl_agents.agents.mab.pursuit import Pursuit  # noqa: F401
from rl_agents.agents.mab.softmax import Softmax  # noqa: F401
from rl_agents.agents.mab.ucbs import UCB, UCB1, UCB2  # noqa: F401



import numpy as np

from rl_agents.agents.mab.base import BaseMAB


class Softmax(BaseMAB):
    """Short summary.

    Parameters
    ----------
    n_arms : type
        Description of parameter `n_arms`.
    temperature : type
        Description of parameter `temperature`.

    Attributes
    ----------
    means : type
        Description of attribute `means`.
    p_arms : type
        Description of attribute `p_arms`.
    trials : type
        Description of attribute `trials`.
    n_arms
    temperature

    """

    def __init__(self, n_arms, temperature):
        self.n_arms = n_arms
        self.means = np.zeros(self.n_arms)
        self.temperature = temperature
        self.p_arms = np.zeros(self.n_arms)
        self.trials = np.zeros(self.n_arms)

    def learn(self, a_idx, reward):
        """Short summary.

        Parameters
        ----------
        a_idx : type
            Description of parameter `a_idx`.
        reward : type
            Description of parameter `reward`.

        Returns
        -------
        type
            Description of returned object.

        """
        self.means[a_idx] = (
            (self.means[a_idx] * self.trials[a_idx]) + reward
        ) / (self.trials[a_idx] + 1)
        self.trials[a_idx] += 1  # add trial

    def predict(self):
        """Short summary.

        Returns
        -------
        type
            Description of returned object.

        """
        e_x = np.exp(self.means / self.temperature)
        self.p_arms = e_x / e_x.sum()
        return np.random.choice(range(self.n_arms), p=self.p_arms)



import numpy as np

from rl_agents.agents.mab.base import BaseMAB


class UCB(BaseMAB):
    r"""MAB Agent following a Upper Confidence Bound policy.

    The UCB selects the action that maximizes the function given by:

    .. math:: f(i) = \mu_i + U_i,

    where  :math:`\mu_i` is the average reward of arm :math:`i`, and
    :math:`U_i` is given by:

    .. math:: U_i = \sqrt{\frac{-\log{p}}{2 N_i} },

    where :math:`N_i` is the number of pulls made to arm  :math:`i`.

    Parameters
    ----------
    n_arms : int
        Number of actions (arms) of the MAB.
    p : float
        Probability of the true value being above the estimate plus the bound.

    Attributes
    ----------
    means : numpy.array(float, ndim=1)
        Vector containing the average reward of each arm.
    trials : numpy.array(float, ndim=1)
        Vector containing the number of trials made to each arm.
    bounds : numpy.array(float, ndim=1)
        Vector containing the upper bounds of each arm.
    t : int
        Total trial counter.

    """

    def __init__(self, n_arms, p):
        self.p = p
        self.n_arms = n_arms
        self.means = np.zeros(self.n_arms)
        self.trials = np.zeros(self.n_arms)
        self.bounds = np.zeros(self.n_arms)
        self.t = 0

    def learn(self, a_idx, reward):
        """Learn from the interaction.

        Update the means, the bounds and the trials.

        Parameters
        ----------
        reward : float
            Reward received from the system after taking action a_idx.
        a_idx : int
            Index of the arm pulled (action taken).

        """
        self.means[a_idx] = (
            (self.means[a_idx] * self.trials[a_idx]) + reward
        ) / (self.trials[a_idx] + 1)
        self.trials[a_idx] += 1  # add trial
        self.bounds[a_idx] = np.sqrt(-np.log(self.p) / 2 * self.trials[a_idx])
        self.t += 1

    def predict(self):
        """Predict next action.

        Pulls each arm once, then chooses the arm that gives the best
        mean + bound.

        Returns
        -------
        int
            Index of chosen action.

        """
        if self.t < self.n_arms:
            return self.t
        return np.argmax(self.means + self.bounds)


class UCB1(BaseMAB):
    """Short summary.

    Parameters
    ----------
    n_arms : type
        Description of parameter `n_arms`.
    c : type
        Description of parameter `c`.

    Attributes
    ----------
    means : type
        Description of attribute `means`.
    trials : type
        Description of attribute `trials`.
    bounds : type
        Description of attribute `bounds`.
    t : type
        Description of attribute `t`.
    n_arms
    c

    """

    def __init__(self, n_arms, c=4):
        self.n_arms = n_arms
        self.means = np.zeros(self.n_arms)
        self.trials = np.zeros(self.n_arms)
        self.bounds = np.zeros(self.n_arms)
        self.c = c
        self.t = 0

    def learn(self, a_idx, reward):
        """Short summary.

        Parameters
        ----------
        a_idx : type
            Description of parameter `a_idx`.
        reward : type
            Description of parameter `reward`.

        Returns
        -------
        type
            Description of returned object.

        """
        self.means[a_idx] = (
            (self.means[a_idx] * self.trials[a_idx]) + reward
        ) / (self.trials[a_idx] + 1)
        self.trials[a_idx] += 1  # add trial
        self.t += 1
        self.bounds[a_idx] = self.c * np.sqrt(
            np.log(self.t) / (self.trials[a_idx])
        )

    def predict(self):
        """Short summary.

        Returns
        -------
        type
            Description of returned object.

        """
        if self.t < self.n_arms:
            return self.t
        return np.argmax(self.means + self.bounds)


class UCB2(BaseMAB):
    """Short summary.

    Parameters
    ----------
    n_arms : type
        Description of parameter `n_arms`.
    alpha : type
        Description of parameter `alpha`.

    Attributes
    ----------
    means : type
        Description of attribute `means`.
    trials : type
        Description of attribute `trials`.
    bounds : type
        Description of attribute `bounds`.
    rj : type
        Description of attribute `rj`.
    t : type
        Description of attribute `t`.
    counter : type
        Description of attribute `counter`.
    current : type
        Description of attribute `current`.
    n_arms
    alpha

    """

    def __init__(self, n_arms, alpha):
        self.n_arms = n_arms
        self.means = np.zeros(self.n_arms)
        self.trials = np.zeros(self.n_arms)
        self.bounds = np.zeros(self.n_arms)
        self.rj = np.zeros(self.n_arms)
        self.alpha = alpha
        self.t = 0
        self.counter = 0
        self.current = 0

    def learn(self, a_idx, reward):
        """Short summary.

        Parameters
        ----------
        a_idx : type
            Description of parameter `a_idx`.
        reward : type
            Description of parameter `reward`.

        Returns
        -------
        type
            Description of returned object.

        """
        self.means[a_idx] = (
            (self.means[a_idx] * self.trials[a_idx]) + reward
        ) / (self.trials[a_idx] + 1)
        self.trials[a_idx] += 1  # add trial
        self.t += 1
        tau = self._tau(self.rj[a_idx])
        self.bounds[a_idx] = np.sqrt(
            (1 + self.alpha) * np.log(np.e * self.t / tau) / (2 * tau)
        )
        self.counter = self._tau(self.rj[a_idx] + 1) - tau
        self.rj[a_idx] += 1

    def predict(self):
        """Short summary.

        Returns
        -------
        type
            Description of returned object.

        """
        if self.t < self.n_arms:
            return self.t
        if self.counter == 0:
            action = np.argmax(self.means + self.bounds)
            self.current = action
            return action
        else:
            self.counter -= 1
            return self.current

    def _tau(self, rj):
        return np.ceil((1 + self.alpha) ** rj)



"""
The :mod:`rl_agents.agents.functions` submodule includes:

* A basic tabular Q-function
* A base class for implementing different functions
"""
from rl_agents.agents.functions.tabular_functions import (
    QMatrixFunction,
    QTableFunction,
)

__all__ = ["QMatrixFunction", "QTableFunction"]



from abc import ABC, abstractmethod


class BaseMAB(ABC):
    """
    A basic Multi-Armed Bandit agent.

    In reinforcement learning, an Agent learns
    by interacting with an Environment.
    Usually, an agent tries to maximize a reward signal.
    It does this by taking "actions" and receiving "rewards",
    and in doing so, learning which action pairs correlate with high rewards.

    An MAB implementation should encapsulate some particular
    reinforcement learning algorihthm.
    """
    @abstractmethod
    def predict(self):
        """Select an action for the current timestep.

        This method allows the agent to do whatever is necessary to select
        an action in a given timestep.
        However, the agent must ultimately return an action (the arm index).

        Returns
        -------
        int
            Index of chosen action.

        """
    @abstractmethod
    def learn(self, a_idx, reward):
        """Learn from the interaction.

        Parameters
        ----------
        a_idx : int
            Index of the arm pulled (action taken).
        reward : float
            Reward received from the system after taking action a_idx.

        """



from abc import ABC, abstractmethod


class BaseQFunction(ABC):
    """
    A basic Q-function RL agents.

    The Q-Function is function that maps a state-action pair to a real value.

    """

    @abstractmethod
    def __call__(self, state, action):
        """Get the Q-value of the state-action pair.

        Parameters
        ----------
        state : type
            The state information
        action : type
            The action information

        Returns
        -------
        float
            Q-Value of the state-action pair.

        """

    @abstractmethod
    def update(self):
        """Update the Q-function.

        """

    @abstractmethod
    def get_values(self, state):
        """Return the Q-Values of all actions in the selected state.

        Parameters
        ----------
        state : type
            State information

        Returns
        -------
        numpy.ndarray(float, ndim=1)
            Q-Values of the actions associated with the state.

        """



import gym
import numpy as np
from gym import spaces
from gym.utils import seeding


class BanditEnv(gym.Env):
    """
    Bandit environment base

    Attributes
    ----------
    arms: int
        Number of arms
    """

    def __init__(self, arms: int):
        self.arms = arms
        self.action_space = spaces.Discrete(self.arms)
        self.observation_space = spaces.Discrete(1)
        self.np_random = None
        self.seed()
        self.reset()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def reset(self):
        raise NotImplementedError

    def step(self, action: int):
        raise NotImplementedError

    def render(self, mode="human", close=False):
        raise NotImplementedError


class BanditKArmedGaussianEnv(BanditEnv):
    """
    K-armed bandit environment.

    This env is mentioned on page 30 of Sutton and Barto's with K=10
    [Reinforcement Learning: An Introduction]
    (https://www.dropbox.com/s/b3psxv2r0ccmf80/book2015oct.pdf?dl=0)

    Actions always pay out
    Mean of payout is pulled from a normal distribution (0, 1) (called q*(a))
    Actual reward is drawn from a normal distribution (q*(a), 1)
    """

    def __init__(self, arms=10):
        self.means = []
        BanditEnv.__init__(self, arms)

    def reset(self):
        self.means = []
        for _ in range(self.arms):
            self.means.append(self.np_random.normal(0, 1))
        self.best_arm = np.argmax(self.means)

    def step(self, action: int):
        assert self.action_space.contains(action)
        reward = self.np_random.normal(self.means[action], 1)
        info = {
            "Optimal": action == self.best_arm,
            "Regret": self.means[self.best_arm] - self.means[action],
        }
        return 0, reward, False, info



"""
The :mod:`rl_agents.agents.policies` submodule includes:

* Epsilon Greedy
* Boltzman (a.k.a Softmax)
"""
from rl_agents.agents.policies.tabular_policies import (
    BoltzmanPolicy,
    EDecreasePolicy,
    EGreedyPolicy,
)

__all__ = ["EGreedyPolicy", "EDecreasePolicy", "BoltzmanPolicy"]



"""
The :mod:`rl_agents.agents` module includes the RL agents
classes and utilities. It includes the MAB variants,
tabular methods and Deep RL.
"""

import rl_agents.agents.mab

from rl_agents.agents.tabular import (  # isort:skip
    QLearningAgent,
    SarsaAgent,
    ExpectedSarsaAgent,
)

__all__ = ["QLearningAgent", "SarsaAgent", "ExpectedSarsaAgent"]



from abc import ABC, abstractmethod


class BaseAgent(ABC):
    """A base RL agent.

    """

    @abstractmethod
    def predict(self, state, eval=False):
        """Predict the next action the agent should take.

        Parameters
        ----------
        state : type
            State information
        eval : bool
            Flag to indicate if the agent is in a test setting (evaluation)

        Returns
        -------
        type
            Action to be taken.

        """

    @abstractmethod
    def learn(self, state, action, reward, next_state):
        """Learn from the interaction.

        Parameters
        ----------
        state : type
            State information.
        action : type
            Action taken in the current step.
        reward : type
            Reward received for the action taken.
        next_state : type
            Next state of the environment.


        """
